import sys  

from .DataManager import datamanager

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from torchvision.models import resnet50, ResNet50_Weights, resnet152
import torchvision

sys.path.insert(0, '../druida_V01/src/')

import os
import copy

from tqdm import tqdm
from torch import optim
import logging 


from torch.utils.tensorboard import SummaryWriter

from .tools import toolkit
from .tools import utils

logging.basicConfig(format="%(asctime)s - %(levelname)s: %(message)s", level=logging.INFO, datefmt="%I:%M:%S")

#Used to implement any prediction process
class Predictor:
    def __init__(self):
        
        self.checkDevice()

    def generate_VAE(self,mean, var, model):
        z_sample = torch.tensor([[mean, var]], dtype=torch.float).to(self.device)
        x_decoded = model.decode(z_sample)
        #image = x_decoded.detach().cpu().reshape(64, 64,3) # reshape vector to 2d array
        
        return x_decoded


    def checkDevice(self):
        self.device = (
        "cuda"
        if torch.cuda.is_available()
        else "mps"
        if torch.backends.mps.is_available()
        else "cpu"
        )

# All training modules for the different algorithms used
class Trainer:
    

    def __init__(self, args):

        self.run_name = args.run_name
        self.learning_rate = args.learning_rate
        self.batch_size = args.batch_size
        self.epochs = args.epochs
        self.workers = args.workers
        self.gpu_number=args.gpu_number

        self.device=self.checkDevice()
        
    def training(self, trainFunction,testFunction, train_dataloader, test_dataloader, model, loss_fn, optimizer):
        acc=0
        acc_test=0
        loss=0
        test_loss=0
        loss_values = []
        test_loss_values = []
        train_acc_hist = []
        test_acc = []

        for t in range(self.epochs):
            dataiter = iter(train_dataloader)
            testdataiter = iter(test_dataloader)


            print(f"Epoch {t+1}\n-------------------------------")
            acc,loss=trainFunction(next(dataiter), model, loss_fn, optimizer, t,acc,loss)
            acc_test,test_loss=testFunction(next(testdataiter), model, loss_fn,len(train_dataloader), t, acc_test,test_loss)

            loss_values.append(loss)
            test_loss_values.append(test_loss)
            train_acc_hist.append(acc)
            test_acc.append(acc_test)
        print("Done!")

        return loss_values,test_loss_values,train_acc_hist, test_acc

    def checkDevice(self):
        self.device = (
        "cuda"
        if torch.cuda.is_available()
        else "mps"
        if torch.backends.mps.is_available()
        else "cpu"
        )


    def multiGPU(self, network):
        print('available Device:'+network.device)
        if (network.device == 'cuda' and (self.gpu_number > 1)):
            network=nn.DataParallel(network,list(range(self.gpu_number)))

        return network
    

    #Traininga  VAE
    def train_VAE(self,args):
        
        utils.setup_logging(self.run_name)

        model = toolkit.VAE().to(self.device)
        optimizer = optim.Adam(model.parameters(), lr=self.learning_rate)   
        dataloader = utils.get_data_denormalize(args.image_size, args.dataset_path,self.batch_size)
        trainingLoss=[]

        logging.info(f"Starting epoch {self.epochs}:")
        
        pbar = tqdm(dataloader)
        model.train()

        for epoch in range(self.epochs):
            overall_loss = 0

            for batch_idx, (x, _) in enumerate(pbar):
                images=x.to(self.device)
                images = (images.clamp(-1, 1) + 1) / 2 #make all values between 0 and 1 
                #images = (x * 255).type(torch.uint8) #valid pixel range
                #images=images.reshape(self.batch_size,-1).to(self.device)
                images = images.view(images.size()[0], (64*64*3)).to(self.device)
                optimizer.zero_grad()

                x_hat, mean, log_var = model(images)
                loss = self.loss_function(images, x_hat, mean, log_var)
                
                overall_loss += loss.item()
                
                loss.backward()
                optimizer.step()
            
            trainingLoss.append(overall_loss)


            print("\tEpoch", epoch + 1, "\tAverage Loss: ", overall_loss/(batch_idx*self.batch_size))
        return trainingLoss,overall_loss, model


    def loss_function(self,x, x_hat, mean, log_var):
     
        reproduction_loss = nn.functional.binary_cross_entropy(x_hat, x, reduction='sum')

        KLD = - 0.5 * torch.sum(1+ log_var - mean.pow(2) - log_var.exp())

        return reproduction_loss + KLD
    

    #N Training an unconditional diffusion model

    def train_DM(self,args):

        utils.setup_logging(self.run_name)

        device = self.device

        dataloader = utils.get_data(args.image_size, args.dataset_path,self.batch_size)

        model = toolkit.UNet(device=self.device,c_in=3, c_out=3, time_dim=256).to(device)

        optimizer = optim.AdamW(model.parameters(), lr=self.learning_rate)

        mse = nn.MSELoss()

        diffusion = Diffusion(img_size=args.image_size, device=device,noise_steps=1000, beta_start=1e-4, beta_end=0.02)

        logger = SummaryWriter(os.path.join("runs", self.run_name))

        l = len(dataloader)

        """this corresponds to algorithm 1 """

        for epoch in range(self.epochs):

            logging.info(f"Starting epoch {epoch}:")

            pbar = tqdm(dataloader)

            for i, (images, _) in enumerate(pbar):
                images = images.to(device)

                t = diffusion.sample_timesteps(images.shape[0]).to(device)

                x_t, noise = diffusion.noise_images(images, t)

                predicted_noise = model(x_t, t)

                loss = mse(noise, predicted_noise)

                optimizer.zero_grad()

                loss.backward()

                optimizer.step()

                pbar.set_postfix(MSE=loss.item())

                logger.add_scalar("MSE", loss.item(), global_step=epoch * l + i)

            sampled_images = diffusion.sample(model, n=images.shape[0])

            utils.save_images(sampled_images, os.path.join("results", self.run_name, f"{epoch}.jpg"))
            torch.save(model.state_dict(), os.path.join("models", self.run_name, f"ckpt.pt"))


    def train_CDM(self,args):

        utils.setup_logging(self.run_name)

        device = self.device
        dataloader = utils.get_data(args.image_size, args.dataset_path,self.batch_size)
        model = toolkit.UNet_conditional(device=self.device,channel_in=3, channel_out=3, time_dim=256,num_classes=args.num_classes).to(device)
        
        optimizer = optim.AdamW(model.parameters(), lr=self.learning_rate)

        mse = nn.MSELoss()

        diffusion = Diffusion(img_size=args.image_size, device=device)
        logger = SummaryWriter(os.path.join("runs", self.run_name))
        l = len(dataloader)

        #Exponential Moving Average
        #We make a copy of the model in order to have a 
        #version in which we use EMA and one without it.

        ema = toolkit.EMA(0.995)

        ema_model = copy.deepcopy(model).eval().requires_grad_(False)

        for epoch in range(self.epochs):

            logging.info(f"Starting epoch {epoch}:")
            pbar = tqdm(dataloader)

            for i, (images, labels) in enumerate(pbar):

                images = images.to(device)
                labels = labels.to(device)


                t = diffusion.sample_timesteps(images.shape[0]).to(device)#Sampling time steps
                x_t, noise = diffusion.noise_images(images, t) #add noise

                #This is for the model to train 10% of the time with no labels
                #Classifier free diffusion guidance
                #Si no hay labels se entrena el clasificador sin condición
                if np.random.random() < 0.1:
                    labels = None


                predicted_noise = model(x_t, t, labels)
                #recibe imagenes con ruido, pasos de tiempo y labels
                loss = mse(noise, predicted_noise)

                optimizer.zero_grad()
                loss.backward()

                optimizer.step()

                #Apply Ema to model

                ema.step_ema(ema_model, model)

                pbar.set_postfix(MSE=loss.item())

                logger.add_scalar("MSE", loss.item(), global_step=epoch * l + i)

            if epoch % 10 == 0:

                labels = torch.arange(args.num_classes).long().to(device)
                print(labels)
                print(len(labels))

                sampled_images = diffusion.sample_cdm(model, n=len(labels), labels=labels)
                ema_sampled_images = diffusion.sample_cdm(ema_model, n=len(labels), labels=labels)
                utils.plot_images(sampled_images)
                utils.save_images(sampled_images, os.path.join("results", self.run_name, f"{epoch}.jpg"))
                utils.save_images(ema_sampled_images, os.path.join("results", self.run_name, f"{epoch}_ema.jpg"))
                torch.save(model.state_dict(), os.path.join("models", self.run_name, f"ckpt.pt"))
                torch.save(ema_model.state_dict(), os.path.join("models", self.run_name, f"ema_ckpt.pt"))
                torch.save(optimizer.state_dict(), os.path.join("models", self.run_name, f"optim.pt"))





#Build a DNN stack 
    
class DNN(nn.Module):

    def __init__(self, layers):
        super(DNN,self).__init__()

        self.checkDevice()
        self.layers = layers
        self.architecture=nn.Sequential()

        for layer in layers:
            self.architecture.add_module(layer['name'],layer['layer'])

        

    def push(self, layer):
        self.layers.append(layer) #each layer must come as dictionary with nn type
        return self.layers
 
    def drop_last(self):
        self.layers.pop() #each layer must come as dictionary with nn type
        return self.layers

    def clear(self):
        self.layers.clear() #each layer must come as dictionary with nn type
        return self.layers
    
    def checkDevice(self):
        self.device = (
        "cuda"
        if torch.cuda.is_available()
        else "mps"
        if torch.backends.mps.is_available()
        else "cpu"
        )
    
    def forward(self, input):

        #{"name":"hidden1","layer":nn.Linear(8,120), "type":"hidden", "index",0}
        self.output =  input
        for layer in self.layers:
            action=layer['layer']
            self.output = action(self.output)
                
                
        return self.output
    


#Build a Generator for GAN stack 

class Generator(nn.Module):
    def __init__(self, ngpu, input_size, mapping_size, channels, leakyRelu_flag ):
        super(Generator, self).__init__()
        

        self.checkDevice()

        self.ngpu = ngpu            

        self.conv1 = nn.ConvTranspose2d(input_size, mapping_size * 8, 6, 1, 0, bias=False)
        self.conv2 = nn.BatchNorm2d(num_features=mapping_size * 8)
        if leakyRelu_flag:
            self.conv3 = nn.ReLU(True)
            self.conv6 = nn.ReLU(True)
            self.conv9 = nn.ReLU(True)
            self.conv12 = nn.ReLU(True)

        else:
            self.conv3 = nn.LeakyReLU(0.2)
            self.conv6 = nn.LeakyReLU(0.2)
            self.conv9 = nn.LeakyReLU(0.2)
            self.conv12 = nn.LeakyReLU(0.2)

        self.conv4 = nn.ConvTranspose2d(mapping_size * 8, mapping_size * 4, 6, 2, 2, bias=False)
        self.conv5 = nn.BatchNorm2d(mapping_size * 4)
        self.conv7 = nn.ConvTranspose2d(mapping_size * 4, mapping_size * 2, 6, 2, 4, bias=False)
        self.conv8 = nn.BatchNorm2d(mapping_size * 2)
        self.conv10 = nn.ConvTranspose2d(mapping_size * 2, mapping_size, 6, 2, 5, bias=False)
        self.conv11 = nn.BatchNorm2d(mapping_size)
        self.conv13 = nn.ConvTranspose2d(mapping_size, channels, 6, 2, 4, bias=False)
        self.conv14 = nn.Tanh()

    def forward(self, input):
        imageOut = input
        imageOut = self.conv1(imageOut)
        imageOut = self.conv2(imageOut)
        imageOut = self.conv3(imageOut)
        imageOut = self.conv4(imageOut)
        imageOut = self.conv5(imageOut)
        imageOut = self.conv6(imageOut)
        imageOut = self.conv7(imageOut)
        imageOut = self.conv8(imageOut)
        imageOut = self.conv9(imageOut)
        imageOut = self.conv10(imageOut)
        imageOut = self.conv11(imageOut)
        imageOut = self.conv12(imageOut)
        imageOut = self.conv13(imageOut)
        imageOut = self.conv14(imageOut)               
        return imageOut

    def checkDevice(self):
        self.device = (
        "cuda"
        if torch.cuda.is_available()
        else "mps"
        if torch.backends.mps.is_available()
        else "cpu"
        )
    

#Build a Discriminator for GAN stack 

class Discriminator(nn.Module):
    def __init__(self,label_length, ngpu=0, image_size=32, discriminator_mapping_size=0, channels=3):
        super(Discriminator, self).__init__()

        
        self.checkDevice()

        self.ngpu = ngpu            
        self.image_size = image_size
        self.channels = channels

        # Habitualmente en el primer ejerciio 
        # label_length era 800

        self.l1 = nn.Linear(label_length, image_size*image_size*channels, bias=False)            
        self.conv1 = nn.Conv2d(2*channels, discriminator_mapping_size, 6, 2, 4, bias=False) 
        self.conv2 = nn.LeakyReLU(0.2, inplace=True)
        self.conv3 = nn.Conv2d(discriminator_mapping_size, discriminator_mapping_size * 2, 6, 2, 5, bias=False)
        self.conv4 = nn.BatchNorm2d(discriminator_mapping_size * 2)
        self.conv5 = nn.LeakyReLU(0.2, inplace=True)
        self.conv6 = nn.Conv2d(discriminator_mapping_size * 2, discriminator_mapping_size * 4, 6, 2, 4, bias=False)
        self.conv7 = nn.BatchNorm2d(discriminator_mapping_size * 4)
        self.conv8 = nn.LeakyReLU(0.2, inplace=True)
        self.conv9 = nn.Conv2d(discriminator_mapping_size * 4, discriminator_mapping_size * 8, 6, 2, 2, bias=False)
        self.conv10 = nn.BatchNorm2d(discriminator_mapping_size * 8)
        self.conv11 = nn.LeakyReLU(0.2, inplace=True)
        self.conv12 = nn.Conv2d(discriminator_mapping_size * 8, 1, 6, 1, 0, bias=False)
        self.conv13 = nn.Sigmoid()


    def forward(self, input, label, b_size):
        x1 = input
        x2 = self.l1(label) #Size must be taken care = 800 in this case
        #the output is imagesize x imagesize x channel
        #hence the need of reshape 

        if self.ngpu == 0 :
        
            x2 = x2.reshape(int(b_size),self.channels,self.image_size,self.image_size) 
        else:
            x2 = x2.reshape(int(b_size/self.ngpu),self.channels,self.image_size,self.image_size) 

        combine = torch.cat((x1,x2),dim=1) # concatenate in a given dimension
        #esto viene del mismo paper sobre los CGAN

        combine = self.conv1(combine) #This conv1 considers 2 x channels from the combine
        combine = self.conv2(combine)
        combine = self.conv3(combine)
        combine = self.conv4(combine)
        combine = self.conv5(combine)
        combine = self.conv6(combine)
        combine = self.conv7(combine)
        combine = self.conv8(combine)
        combine = self.conv9(combine)
        combine = self.conv10(combine)
        combine = self.conv11(combine)
        combine = self.conv12(combine)
        combine = self.conv13(combine)
        return combine

    def checkDevice(self):
        self.device = (
        "cuda"
        if torch.cuda.is_available()
        else "mps"
        if torch.backends.mps.is_available()
        else "cpu"
        )
    

#Build a Diffusion  model pipeline

# http://layer-calc.com/
class Generator_V2(nn.Module):
    def __init__(self, ngpu,spectra_input_size,latent_size, mapping_size,initial_depth, channels, leakyRelu_flag ):
        super(Generator_V2, self).__init__()
        

        self.checkDevice()
        self.spectra_len=spectra_input_size
        self.latent_len=latent_size
        self.ngpu = ngpu            
        # 512 canales de entrada
        self.depth=initial_depth#512
        self.l1 = nn.Linear(self.spectra_len, 4*4*self.depth, bias=False)            
        self.l2 = nn.Linear(self.latent_len, 4*4*self.depth, bias=False)            

        self.conv1 = nn.ConvTranspose2d(self.depth*2, mapping_size * 8, 5, 1, 0, bias=False)
        self.conv2 = nn.BatchNorm2d(num_features=mapping_size * 8)
        if leakyRelu_flag:
            self.conv3 = nn.ReLU(True)
            self.conv6 = nn.ReLU(True)
            self.conv9 = nn.ReLU(True)
            self.conv12 = nn.ReLU(True)

        else:
            self.conv3 = nn.LeakyReLU(0.2)
            self.conv6 = nn.LeakyReLU(0.2)
            self.conv9 = nn.LeakyReLU(0.2)
            self.conv12 = nn.LeakyReLU(0.2)

        self.conv4 = nn.ConvTranspose2d(mapping_size * 8, mapping_size * 4, 6, 2, 2, bias=False)
        self.conv5 = nn.BatchNorm2d(mapping_size * 4)
        self.conv7 = nn.ConvTranspose2d(mapping_size * 4, mapping_size * 2, 6, 2, 2, bias=False)
        self.conv8 = nn.BatchNorm2d(mapping_size * 2)
        self.conv10 = nn.ConvTranspose2d(mapping_size * 2, mapping_size, 6, 2, 2, bias=False)
        self.conv11 = nn.BatchNorm2d(mapping_size)
        self.conv13 = nn.ConvTranspose2d(mapping_size, channels, 5, 1, 2, bias=False)
        self.conv14 = nn.Tanh()

    def forward(self, input, latent, b_size):

        x_condspectra = self.l1(input) #Size must be taken care = 800 in this case
        x_latent = self.l2(latent) #Size must be taken care = 800 in this case

        if self.ngpu == 0 :
        
            x_condspectra = x_condspectra.reshape(int(b_size),self.depth,4,4) 
            x_latent = x_latent.reshape(int(b_size),self.depth,4,4) 

        else:
            x_condspectra = x_condspectra.reshape(int(b_size/self.ngpu),self.depth,4,4) 
            x_latent = x_latent.reshape(int(b_size/self.ngpu),self.depth,4,4) 

        combine = torch.cat((x_latent,x_condspectra),dim=1) # concatenate in a given dimension
        #esto viene d

        imageOut = combine

        imageOut = self.conv1(imageOut)
        imageOut = self.conv2(imageOut)
        imageOut = self.conv3(imageOut)
        imageOut = self.conv4(imageOut)
        imageOut = self.conv5(imageOut)
        imageOut = self.conv6(imageOut)
        imageOut = self.conv7(imageOut)
        imageOut = self.conv8(imageOut)
        imageOut = self.conv9(imageOut)
        imageOut = self.conv10(imageOut)
        imageOut = self.conv11(imageOut)
        imageOut = self.conv12(imageOut)
        imageOut = self.conv13(imageOut)
        imageOut = self.conv14(imageOut)               
        return imageOut

    def checkDevice(self):
        self.device = (
        "cuda"
        if torch.cuda.is_available()
        else "mps"
        if torch.backends.mps.is_available()
        else "cpu"
        )
    

#Build a Discriminator for GAN stack 

class Discriminator_V2(nn.Module):
    def __init__(self,label_length, ngpu=0, image_size=32, discriminator_mapping_size=0, channels=3):
        super(Discriminator, self).__init__()

        
        self.checkDevice()

        self.ngpu = ngpu            
        self.image_size = image_size
        self.channels = channels

        # Habitualmente en el primer ejerciio 
        # label_length era 800

        self.l1 = nn.Linear(label_length, image_size*image_size*channels, bias=False)            
        self.conv1 = nn.Conv2d(2*channels, discriminator_mapping_size, 6, 2, 4, bias=False) 
        self.conv2 = nn.LeakyReLU(0.2, inplace=True)
        self.conv3 = nn.Conv2d(discriminator_mapping_size, discriminator_mapping_size * 2, 6, 2, 5, bias=False)
        self.conv4 = nn.BatchNorm2d(discriminator_mapping_size * 2)
        self.conv5 = nn.LeakyReLU(0.2, inplace=True)
        self.conv6 = nn.Conv2d(discriminator_mapping_size * 2, discriminator_mapping_size * 4, 6, 2, 4, bias=False)
        self.conv7 = nn.BatchNorm2d(discriminator_mapping_size * 4)
        self.conv8 = nn.LeakyReLU(0.2, inplace=True)
        self.conv9 = nn.Conv2d(discriminator_mapping_size * 4, discriminator_mapping_size * 8, 6, 2, 2, bias=False)
        self.conv10 = nn.BatchNorm2d(discriminator_mapping_size * 8)
        self.conv11 = nn.LeakyReLU(0.2, inplace=True)
        self.conv12 = nn.Conv2d(discriminator_mapping_size * 8, 1, 6, 1, 0, bias=False)
        self.conv13 = nn.Sigmoid()


    def forward(self, input, label, b_size):
        x1 = input
        x2 = self.l1(label) #Size must be taken care = 800 in this case
        #the output is imagesize x imagesize x channel
        #hence the need of reshape 

        if self.ngpu == 0 :
        
            x2 = x2.reshape(int(b_size),self.channels,self.image_size,self.image_size) 
        else:
            x2 = x2.reshape(int(b_size/self.ngpu),self.channels,self.image_size,self.image_size) 

        combine = torch.cat((x1,x2),dim=1) # concatenate in a given dimension
        #esto viene del mismo paper sobre los CGAN

        combine = self.conv1(combine) #This conv1 considers 2 x channels from the combine
        combine = self.conv2(combine)
        combine = self.conv3(combine)
        combine = self.conv4(combine)
        combine = self.conv5(combine)
        combine = self.conv6(combine)
        combine = self.conv7(combine)
        combine = self.conv8(combine)
        combine = self.conv9(combine)
        combine = self.conv10(combine)
        combine = self.conv11(combine)
        combine = self.conv12(combine)
        combine = self.conv13(combine)
        return combine

    def checkDevice(self):
        self.device = (
        "cuda"
        if torch.cuda.is_available()
        else "mps"
        if torch.backends.mps.is_available()
        else "cpu"
        )
    


class Predictor_CNN(nn.Module):
    def __init__(self, cond_input_size, ngpu=0, image_size=512 , output_size=0, channels=3,features_num=1000, hiden_num=5000, dropout=0.2, Y_prediction_size=601):
        super(Predictor_CNN, self).__init__()

        
        self.checkDevice()

        self.ngpu = ngpu            
        self.image_size = image_size
        self.channels = channels
        self.features_num = features_num
        self.dropout=dropout


        self.l1 = nn.Linear(cond_input_size, image_size*image_size*channels, bias=False)           
        self.conv1 = nn.Conv2d(2*channels, output_size, 6, 2, 4, bias=False) 
        self.conv2 = nn.LeakyReLU(0.2, inplace=True)
        self.conv3 = nn.Conv2d(output_size, output_size * 2, 6, 2, 5, bias=False)
        self.conv4 = nn.BatchNorm2d(output_size * 2)
        self.conv5 = nn.LeakyReLU(0.2, inplace=True)
        self.conv6 = nn.Conv2d(output_size * 2, output_size * 4, 6, 2, 4, bias=False)
        self.conv7 = nn.BatchNorm2d(output_size * 4)
        self.conv8 = nn.LeakyReLU(0.2, inplace=True)
        self.conv9 = nn.Conv2d(output_size * 4, output_size * 8, 6, 2, 2, bias=False)
        self.conv10 = nn.BatchNorm2d(output_size * 8)
        self.conv11 = nn.LeakyReLU(0.2, inplace=True)
        self.conv12 = nn.Conv2d(output_size * 8, output_size * 16, 2, 1, 0, bias=False)

        
        #Here we apply a flattening
        #10368
        #139392
        if image_size==128:
            self.l2 = nn.Linear(10368, hiden_num, bias=False)           

        elif image_size==512:
            self.l2 = nn.Linear(139392, hiden_num, bias=False)           
        elif image_size==64:
            self.l2 = nn.Linear(3200, hiden_num, bias=False)           
        
        #self.l3 = nn.Linear(hiden_num, features_num, bias=False)           
        self.dropout2 = nn.Dropout(dropout)
        self.l4 = nn.Linear(hiden_num, hiden_num, bias=False)           
        self.l5 = nn.Linear(hiden_num, hiden_num, bias=False)           
        self.l6 = nn.Linear(hiden_num, Y_prediction_size, bias=False)           


    def forward(self, input_, conditioning, b_size):
        x1 = input_
        x2 = self.l1(conditioning) #Size must be taken care = 800 in this case
        #the output is imagesize x imagesize x channel
        #hence the need of reshape 
        #print(x2.shape)
        if self.ngpu == 0 :
        
            x2 = x2.reshape(int(b_size),self.channels,self.image_size,self.image_size) 
        else:
            x2 = x2.reshape(int(b_size/self.ngpu),self.channels,self.image_size,self.image_size) 
        
        #print(x2.shape)

        combine = torch.cat((x1,x2),dim=1) # concatenate in a given dimension
     
        outmap_min, _ = torch.min(combine, dim=1, keepdim=True)
        outmap_max, _ = torch.max(combine, dim=1, keepdim=True)
        combine = (combine - outmap_min) / (outmap_max - outmap_min) 

        combine = self.conv1(combine) #This conv1 considers 2 x channels from the combine
        combine = self.conv2(combine)
        combine = self.conv3(combine)
        combine = self.conv4(combine)
        combine = self.conv5(combine)
        
        combine = self.conv6(combine)
        combine = self.conv7(combine)
        combine = self.conv8(combine)
        combine = self.conv9(combine)
        combine = self.conv10(combine)
        combine = self.conv11(combine)
        combine = self.conv12(combine)

        """Change between conv to linear layers"""

        combine =combine.view(combine.size(0), -1)
        combine = self.l2(combine)
        #combine = self.l3(combine)

        combine = self.dropout2(combine)

        combine = self.l4(combine)
        
        combine = self.dropout2(combine)

        combine = self.l5(combine)
        combine = self.l6(combine)

        #a Final softmax
        #combine = F.softmax(combine, dim=1)


        return combine

    def checkDevice(self):
        self.device = (
        "cuda"
        if torch.cuda.is_available()
        else "mps"
        if torch.backends.mps.is_available()
        else "cpu"
        )
    

class Predictor_RESNET(nn.Module):
    def __init__(self,resnet_arch,conditional, cond_input_size,cond_channels, ngpu=0, image_size=512 , output_size=0, channels=3,features_num=1000, hiden_num=5000, dropout=0.2, Y_prediction_size=601):
        super(Predictor_RESNET, self).__init__()

        
        self.checkDevice()

        self.ngpu = ngpu            
        self.image_size = image_size
        self.channels = channels
        self.features_num = features_num
        self.dropout=dropout
        self.cond_channels=cond_channels

        self.l1 = nn.Linear(cond_input_size, image_size*image_size*cond_channels, bias=False)           

        #weights = ResNet50_Weights.DEFAULT
        if resnet_arch=="resnet50":
            self.model = resnet50(pretrained=False)
        else:
            self.model = resnet152(pretrained=False)



        self.conditional=conditional
        num_filters = self.model.conv1.out_channels   
        kernel_size = self.model.conv1.kernel_size
        stride = self.model.conv1.stride
        padding = self.model.conv1.padding

        if self.conditional==True:
            conv1 = torch.nn.Conv2d(self.cond_channels+self.channels, num_filters, kernel_size=kernel_size, stride=stride, padding=padding)
            # Initialize the new conv1 layer's weights by averaging the pretrained weights across the channel dimension
            #original_weights = self.model.conv1.weight.data.mean(dim=1, keepdim=True)
            # Expand the averaged weights to the number of input channels of the new dataset
            #conv1.weight.data = original_weights.repeat(1, 6, 1, 1)
        else:
            conv1 = torch.nn.Conv2d(self.channels, num_filters, kernel_size=kernel_size, stride=stride, padding=padding)
            # Initialize the new conv1 layer's weights by averaging the pretrained weights across the channel dimension
            #original_weights = self.model.conv1.weight.data.mean(dim=1, keepdim=True)
            # Expand the averaged weights to the number of input channels of the new dataset
            #conv1.weight.data = original_weights.repeat(1, 3, 1, 1)        # Substitute the FC output layer

        self.model.conv1 = conv1

        self.linear = nn.Sequential(
                nn.Linear(self.model.fc.in_features, self.model.fc.in_features, bias=False),
                nn.Linear(self.model.fc.in_features, features_num, bias=False),
                nn.Linear(features_num, hiden_num, bias=False),
                nn.Dropout(dropout),
                nn.Linear(hiden_num, Y_prediction_size, bias=False)           
            )
        
        self.model.fc =self.linear
        
        #torch.nn.init.xavier_uniform_(self.model.fc.weight)


    def forward(self, input_, conditioning, b_size):
        x1 = input_
        #the output is imagesize x imagesize x channel
        #hence the need of reshape 
        #print(x2.shape)
        #We can tray as many channels as prefered. In this case I will try 1 channel
        num_channel=self.cond_channels

        if self.conditional:
            
            x2 = self.l1(conditioning) #Size must be taken care = 800 in this case
            m = nn.Tanh()
            x2 = m(x2)

            if self.ngpu == 0 :
            
                x2 = x2.reshape(int(b_size),num_channel,self.image_size,self.image_size) 
                
            else:
                x2 = x2.reshape(int(b_size/self.ngpu),num_channel,self.image_size,self.image_size)

            # if self.cond_channels==3:
            #     x2 = torchvision.transforms.Normalize([0.6, ], [0.3, ],[0.8,])(x2)
            # else:
            #     x2 = torchvision.transforms.Normalize([0.2, ], [0.1, ],[0.3,])(x2)


            combine = torch.cat((x1,x2),dim=1) # concatenate in a given dimension
        
            #outmap_min, _ = torch.min(combine, dim=1, keepdim=True)
            #outmap_max, _ = torch.max(combine, dim=1, keepdim=True)
            #combine = (combine - outmap_min) / (outmap_max - outmap_min) 

            combine = self.model(combine) #This conv1 considers 2 x channels from the combine
        
            """Change between conv to linear layers"""
            return combine
        else:
            return self.model(x1)

    def checkDevice(self):
        self.device = (
        "cuda"
        if torch.cuda.is_available()
        else "mps"
        if torch.backends.mps.is_available()
        else "cpu"
        )
    


class Diffusion:

    def __init__(self,device, noise_steps=1000, beta_start=1e-4, beta_end=0.02, img_size=64 ):
        self.noise_steps = noise_steps
        self.beta_start = beta_start
        self.beta_end = beta_end
        self.img_size = img_size
        self.device = device

        self.beta = self.prepare_noise_schedule().to(device) #prepares noise for every step
        self.alpha = 1. - self.beta #calc alpha for every beta step
        self.alpha_hat = torch.cumprod(self.alpha, dim=0) #get alpha hat which is the producof all alphas

    def prepare_noise_schedule(self):
        #this creates a linear distribution of beta along the time steps
        """pending! implement cosine beta distribution"""

        return torch.linspace(self.beta_start, self.beta_end, self.noise_steps)

    def noise_images(self, x, t):

        #X_t sample in any arbitrary time
        #one way to do this is by adding noise step by step 
        #the other way is to doit all at once
        sqrt_alpha_hat = torch.sqrt(self.alpha_hat[t])[:, None, None, None]
        sqrt_one_minus_alpha_hat = torch.sqrt(1 - self.alpha_hat[t])[:, None, None, None]

        #noise
        Ɛ = torch.randn_like(x)
        return sqrt_alpha_hat * x + sqrt_one_minus_alpha_hat * Ɛ, Ɛ

    def sample_timesteps(self, n):

        return torch.randint(low=1, high=self.noise_steps, size=(n,))



    """This corresponds to Algorithm Sampling"""
    """We need to sample in reverse direction
    This means from a X_t value, to obtain the previous step.
    And finally obtain X_0"""

    def sample(self, model, n):
        logging.info(f"Sampling {n} new images....")

        model.eval() #Model used for sampling

        with torch.no_grad():

            """Generate n images with 3 channels."""
            x = torch.randn((n, 3, self.img_size, self.img_size)).to(self.device) #create initial images with random noise

            for i in tqdm(reversed(range(1, self.noise_steps)), position=0): #loop over time steps

                t = (torch.ones(n) * i).long().to(self.device)

                predicted_noise = model(x, t)
                alpha = self.alpha[t][:, None, None, None]
                alpha_hat = self.alpha_hat[t][:, None, None, None]
                beta = self.beta[t][:, None, None, None]

                #Excluding the first images?!
                if i > 1:
                    noise = torch.randn_like(x) #Returns a tensor with the same size as input that is filled with random numbers from a normal distribution with mean 0 and variance 1.
                else:
                    noise = torch.zeros_like(x)#Returns a tensor with the same size as input that is filled with random numbers from a normal distribution with mean 0 and variance 1.

                #This is the cool formula with epsilon namely the noise perturbation
                #Removing a little noise in each step
                x = 1 / torch.sqrt(alpha) * (x - ((1 - alpha) / (torch.sqrt(1 - alpha_hat))) * predicted_noise) + torch.sqrt(beta) * noise

        model.train()
        x = (x.clamp(-1, 1) + 1) / 2 #make all values between 0 and 1 
        x = (x * 255).type(torch.uint8) #valid pixel range
        return x
    

    #Sample conditional diffusion
    def sample_cdm(self, model, n, labels, cfg_scale=3):
        logging.info(f"Sampling {n} new images....")
        model.eval()
        with torch.no_grad():
            x = torch.randn((n, 3, self.img_size, self.img_size)).to(self.device)
            for i in tqdm(reversed(range(1, self.noise_steps)), position=0):
                t = (torch.ones(n) * i).long().to(self.device)
                predicted_noise = model(x, t, labels)

                #This is to enable CFG conditional fre guidance
                #torch lerp interpolates the unconditional predicted noise

                if cfg_scale > 0:
                    uncond_predicted_noise = model(x, t, None)
                    predicted_noise = torch.lerp(uncond_predicted_noise, predicted_noise, cfg_scale)

                alpha = self.alpha[t][:, None, None, None]
                alpha_hat = self.alpha_hat[t][:, None, None, None]
                beta = self.beta[t][:, None, None, None]
                if i > 1:
                    noise = torch.randn_like(x)
                else:
                    noise = torch.zeros_like(x)
                x = 1 / torch.sqrt(alpha) * (x - ((1 - alpha) / (torch.sqrt(1 - alpha_hat))) * predicted_noise) + torch.sqrt(beta) * noise
        model.train()
        x = (x.clamp(-1, 1) + 1) / 2
        x = (x * 255).type(torch.uint8)
        return x



"""Stable difussion stack"""
class DiffusionWrapper(nn.Module):
    def __init__(self, diffusion_model: toolkit.UNetModel):
        super().__init__()
        self.diffusion_model = diffusion_model

    def forward(self, x: torch.Tensor, time_steps: torch.Tensor, context: torch.Tensor):
        return self.diffusion_model(x, time_steps, context)

class StableDiffusion(nn.Module):
    
    model: DiffusionWrapper
    first_stage_model: toolkit.Autoencoder
    cond_stage_model: toolkit.CLIPTextEmbedder
    
    def __init__(self,
                 unet_model: toolkit.UNetModel ,
                 autoencoder: toolkit.Autoencoder,
                 clip_embedder: toolkit.CLIPTextEmbedder,
                 latent_scaling_factor: float,
                 n_steps: int,
                 linear_start: float,
                 linear_end: float,
                 ):
        # unet_model is the U-Net that predicts noise ϵ_cond(x_t,c), in latent space
        # autoencoder is the AutoEncoder
        # clip_embedder is the CLIP embeddings generator
        # latent_scaling_factor is the scaling factor for the latent space. The encodings of the autoencoder are scaled by this before feeding into the U-Net.
        # n_steps is the number of diffusion steps T.
        # linear_start is the start of the β schedule.
        # linear_end is the end of the β schedule.
        super().__init__()
        self.model = DiffusionWrapper(unet_model)

        self.first_stage_model = autoencoder
        self.latent_scaling_factor = latent_scaling_factor
        self.cond_stage_model = clip_embedder 

        self.n_steps = n_steps #Number of steps T

        #ß scheduler
        beta = torch.linspace(linear_start ** 0.5, linear_end ** 0.5, n_steps, dtype=torch.float64) ** 2
        self.beta = nn.Parameter(beta.to(torch.float32), requires_grad=False)

        alpha = 1. - beta
        alpha_bar = torch.cumprod(alpha, dim=0)
        self.alpha_bar = nn.Parameter(alpha_bar.to(torch.float32), requires_grad=False)
        
    @property
    def device(self):
        return next(iter(self.model.parameters())).device
    
    def get_text_conditioning(self, prompts: list[str]):    
        
        return self.cond_stage_model(prompts)
    
   #Get scaled latent space representation of the image
    #The encoder output is a distribution. We sample from that and multiply by the scaling factor.
    def autoencoder_encode(self, image: torch.Tensor):
        return self.latent_scaling_factor * self.first_stage_model.encode(image).sample()
    
    #Get image from the latent representation
    #   We scale down by the scaling factor and then decode.
    def autoencoder_decode(self, z: torch.Tensor):
        return self.first_stage_model.decode(z / self.latent_scaling_factor)
    
    #Predict noise given the latent representation xt, time step t, and the conditioning context c.
    def forward(self, x: torch.Tensor, t: torch.Tensor, context: torch.Tensor):
        return self.model(x, t, context)



"""Vision Transformer"""
"""PRE LN architecture"""

"""PRE LN architecture"""

class AttentionBlock(nn.Module):
    def __init__(self, embed_dim, hidden_dim, num_heads, dropout=0.0):
        """Attention Block.

        Args:
            embed_dim: Dimensionality of input and attention feature vectors
            hidden_dim: Dimensionality of hidden layer in feed-forward network
                         (usually 2-4x larger than embed_dim)
            num_heads: Number of heads to use in the Multi-Head Attention block
            dropout: Amount of dropout to apply in the feed-forward network
        """
        super().__init__()

        self.layer_norm_1 = nn.LayerNorm(embed_dim)
        self.attn = nn.MultiheadAttention(embed_dim, num_heads)
        self.layer_norm_2 = nn.LayerNorm(embed_dim)
        self.linear = nn.Sequential(
            nn.Linear(embed_dim, hidden_dim),
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Linear(hidden_dim, embed_dim),
            nn.Dropout(dropout),
        )

    def forward(self, x):
        inp_x = self.layer_norm_1(x)
        x = x + self.attn(inp_x, inp_x, inp_x)[0]
        x = self.layer_norm_2(x)

        x = x + self.linear(x)
        return x
    
class VisionTransformer(nn.Module):
    def __init__(
        self,
       **kwargs
    ):

        """Vision Transformer.

        Args:
            embed_dim: Dimensionality of the input feature vectors to the Transformer
            hidden_dim: Dimensionality of the hidden layer in the feed-forward networks
                         within the Transformer usually 2-4x larger than embed_dim
            num_channels: Number of channels of the input (3 for RGB)
            num_heads: Number of heads to use in the Multi-Head Attention block
            num_layers: Number of layers to use in the Transformer
            num_classes: Number of classes to predict
            patch_size: Number of pixels that the patches have per dimension
            num_patches: Maximum number of patches an image can have
            dropout: Amount of dropout to apply in the feed-forward network and
                      on the input encoding
        """
        super().__init__()
        
        
        self.batch_size=kwargs["batch_size"]
        embed_dim=kwargs["embed_dim"]
        hidden_dim=kwargs["hidden_dim"]
        self.num_channels=kwargs["num_channels"]
        self.cond_num_channels=kwargs["con_num_channels"]
        num_heads=kwargs["num_heads"]
        num_layers=kwargs["num_layers"]
        num_classes=kwargs["num_classes"]
        patch_size=kwargs["patch_size"]
        num_patches=kwargs["num_patches"]
        dropout=kwargs["dropout"]
        self.image_size=kwargs["image_size"]
        self.regression=kwargs["regression"]

        self.patch_size = patch_size
        self.ngpu=0

        if kwargs['conditionalIn'] == True :

            print('conditioned')
            self.conditionalIn=kwargs["conditionalIn"]
            self.conditionalLen=kwargs["conditionalLen"]

            self.l1 = nn.Linear(self.conditionalLen, self.image_size*self.image_size*self.cond_num_channels, bias=False)           


            # Layers/Networks
            
            self.input_layer = nn.Linear((self.cond_num_channels+self.num_channels) * (patch_size**2), 2*embed_dim)
            self.transformer = nn.Sequential(
                *(AttentionBlock(2*embed_dim, hidden_dim, num_heads, dropout=dropout) for _ in range(num_layers))
            )


            # If regression or classification
            if self.regression:
                self.mlp_head = nn.Sequential(
                    nn.LayerNorm(2*embed_dim),
                    nn.Linear(2*embed_dim, 2*embed_dim, bias=False),
                    nn.Dropout(dropout),
                    nn.Linear(2*embed_dim, num_classes, bias=False) 
                )

            else:
                self.mlp_head = nn.Sequential(nn.LayerNorm(2*embed_dim), nn.Linear(2*embed_dim, num_classes))

            self.dropout = nn.Dropout(dropout)

            # Parameters/Embeddings
            #self.cls_token = nn.Parameter(torch.randn(1, 1, embed_dim))
            
            self.cls_token = nn.Parameter(torch.randn(self.batch_size, 1, (self.cond_num_channels+self.num_channels) * (patch_size)**2))
            
            
            self.pos_embedding = nn.Parameter(torch.randn(self.batch_size, 1 + num_patches, 2*embed_dim))



        else:
            # Layers/Networks
            
            self.input_layer = nn.Linear(self.num_channels * (patch_size**2), embed_dim)
            self.transformer = nn.Sequential(
                *(AttentionBlock(embed_dim, hidden_dim, num_heads, dropout=dropout) for _ in range(num_layers))
            )

            # If regression or classification

            if self.regression:
                self.mlp_head = nn.Sequential(
                    nn.LayerNorm(2*embed_dim),
                    nn.Linear(2*embed_dim, 2*embed_dim, bias=False),
                    nn.Dropout(dropout),
                    nn.Linear(2*embed_dim, num_classes, bias=False) 
                )

            else:
                self.mlp_head = nn.Sequential(nn.LayerNorm(embed_dim),
                                              nn.Linear(embed_dim, embed_dim),
                                              nn.Linear(embed_dim, num_classes))
            
            self.dropout = nn.Dropout(dropout)

            # Parameters/Embeddings
            #self.cls_token = nn.Parameter(torch.randn(1, 1, embed_dim))
            
            self.cls_token = nn.Parameter(torch.randn(self.batch_size, 1, self.num_channels * patch_size**2))
            
            
            self.pos_embedding = nn.Parameter(torch.randn(self.batch_size, 1 + num_patches, embed_dim))

        #self.apply(self._init_weights) 

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            nn.init.trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def patchify(self,x, patch_size, flatten_channels):
        """
        Args:
            x: Tensor representing the image of shape [B, C, H, W]
            patch_size: Number of pixels per dimension of the patches (integer)
            flatten_channels: If True, the patches will be returned in a flattened format
                            as a feature vector instead of a image grid.
        """
            
        B, C, H, W = x.shape
        x = x.reshape(B, C, H // patch_size, patch_size, W // patch_size, patch_size)
        x = x.permute(0, 2, 4, 1, 3, 5)  # [B, H', W', C, p_H, p_W]
        x = x.flatten(1, 2)  # [B, H'*W', C, p_H, p_W]
        if flatten_channels:
            x = x.flatten(2, 4)  # [B, H'*W', C*p_H*p_W]
            """Batches, number of patches per image, each flattened patch size including  channel"""
        return x  
        
    def forward(self, x, condition=None):
        # Preprocess input
        #x = img_to_patch(x, self.patch_size)
        if condition !=None:
 
            x2 = self.l1(condition) 

            if self.ngpu == 0 :
            
                x2 = x2.reshape(int(self.batch_size),self.cond_num_channels,self.image_size,self.image_size) 
            else:
                x2 = x2.reshape(int(self.batch_size/self.ngpu),self.cond_num_channels,self.image_size,self.image_size) 
            
            x2 = torchvision.transforms.Normalize([0.15, ], [.1, ],[0.4,])(x2)

            combine = torch.cat((x,x2),dim=1) # concatenate in a given dimension
            
 
            outmap_min, _ = torch.min(combine, dim=1, keepdim=True)
            outmap_max, _ = torch.max(combine, dim=1, keepdim=True)
            x = (combine - outmap_min) / (outmap_max - outmap_min) 
    
        B, T, _, _ = x.shape


        x = self.patchify(x, patch_size=self.patch_size, flatten_channels=True)
        x = self.input_layer(x)

        # Add CLS token and positional encoding
        cls_token = self.cls_token #.repeat(B, 1, 1)
        #print("CLS Token  ",cls_token.shape)

        x = torch.cat([cls_token, x], dim=1)

        pos_embed=self.pos_embedding
        #print("Positional Embedd ",pos_embed.shape)
        
        x = x + pos_embed
        #print("x ", x.shape)
        # Apply Transforrmer
        x = self.dropout(x)
        x = x.transpose(0, 1)
        x = self.transformer(x)

        
        # Perform classification prediction
        cls = x[0]

        # If regression or classification
        # with if in case necessary in the future
        if self.regression:

            out = self.mlp_head(cls)

        else:
            out = self.mlp_head(cls)

        return out
    
    



class MetaSurface:

    def __init__(self, size, grid_size, atom_grid, **kwargs  ):
        self.size = size
        self.grid_size = grid_size
        self.grid_atom=atom_grid

        pass

    def checkKwargArgument(self):
        pass
    


    def create(self):
        pass

    def checkGrid(self):
        pass






import torch
import torch.nn as nn
import torch.nn.functional as F

from .unetAttention import SpatialTransformer

from typing import List

from torch import nn
from transformers import CLIPTokenizer, CLIPTextModel
"""We need to develop an encoder-bottle neck- decoder architecture"""

#Exponential Moving Average
class EMA:
    def __init__(self, beta):
        super().__init__()
        self.beta = beta
        self.step = 0

    def update_model_average(self, ma_model, current_model):

        for current_params, ma_params in zip(current_model.parameters(), ma_model.parameters()):
            old_weight, up_weight = ma_params.data, current_params.data
            ma_params.data = self.update_average(old_weight, up_weight)

    def update_average(self, old, new):
        if old is None:
            return new
        return old * self.beta + (1 - self.beta) * new

    def step_ema(self, ema_model, model, step_start_ema=2000):
        if self.step < step_start_ema:
            self.reset_parameters(ema_model, model)
            self.step += 1
            #increasing through iterations
            return
        
        self.update_model_average(ema_model, model)
        self.step += 1

    def reset_parameters(self, ema_model, model):

        #replicate model params into ema-model
        ema_model.load_state_dict(model.state_dict()) 
        #state_dict is simply a Python dictionary object that maps each layer to its parameter tensor


class UNet(nn.Module):
    def __init__(self,device, c_in=3, c_out=3, time_dim=256):
        super().__init__()
        self.device = device
        self.time_dim = time_dim

        #here we find the full
        #Unet pipeline
        #every step 

        #Encoder
        self.inc = DoubleConv(c_in, 64)
        self.down1 = DownScaler(64, 128) #Reduce image resolution by 2
        self.sa1 = SelfAttention(128, 32) #channel dimension and current resultuion. Output from downscaler is 32x32
        self.down2 = DownScaler(128, 256)
        self.sa2 = SelfAttention(256, 16)
        self.down3 = DownScaler(256, 256)
        self.sa3 = SelfAttention(256, 8)

        #Bottle neck

        self.bot1 = DoubleConv(256, 512)
        self.bot2 = DoubleConv(512, 512)
        self.bot3 = DoubleConv(512, 256)
  
        #Decoder

        self.up1 = UpScaler(512, 128)
        self.sa4 = SelfAttention(128, 16)
        self.up2 = UpScaler(256, 64)
        self.sa5 = SelfAttention(64, 32)
        self.up3 = UpScaler(128, 64)
        self.sa6 = SelfAttention(64, 64)
        self.outc = nn.Conv2d(64, c_out, kernel_size=1) #Projecting out the image


    ##Postional encoding
    def pos_encoding(self, t, channels):
        inv_freq = 1.0 / (
            10000
            ** (torch.arange(0, channels, 2, device=self.device).float() / channels)
        )
        pos_enc_a = torch.sin(t.repeat(1, channels // 2) * inv_freq)
        pos_enc_b = torch.cos(t.repeat(1, channels // 2) * inv_freq)
        pos_enc = torch.cat([pos_enc_a, pos_enc_b], dim=-1)
        return pos_enc.to(self.device)

    def forward(self, x, t): #noised images and time steps
        t = t.unsqueeze(-1).type(torch.float)
        t = self.pos_encoding(t, self.time_dim)
        """we are adding positional information for the time steps"""

        x1 = self.inc(x)
        x2 = self.down1(x1, t)
        x2 = self.sa1(x2)
        x3 = self.down2(x2, t)
        x3 = self.sa2(x3)
        x4 = self.down3(x3, t)
        x4 = self.sa3(x4)


        x4 = self.bot1(x4)
        x4 = self.bot2(x4)
        x4 = self.bot3(x4)

        #Here we have to understand why x4 and x3?!
        #look to UpScaler forwards pass
        x = self.up1(x4, x3, t)
        x = self.sa4(x)
        x = self.up2(x, x2, t)
        x = self.sa5(x)
        x = self.up3(x, x1, t)
        x = self.sa6(x)
        output = self.outc(x)
        return output
    

class UNet_conditional(nn.Module):
    def __init__(self,device,  channel_in=3, channel_out=3, time_dim=256, num_classes=None):
        super().__init__()
        
        self.device = device
        self.time_dim = time_dim
        self.inc = DoubleConv(channel_in, 64)
        self.down1 = DownScaler(64, 128)
        self.sa1 = SelfAttention(128, 32)
        self.down2 = DownScaler(128, 256)
        self.sa2 = SelfAttention(256, 16)
        self.down3 = DownScaler(256, 256)
        self.sa3 = SelfAttention(256, 8)

        self.bot1 = DoubleConv(256, 512)
        self.bot2 = DoubleConv(512, 512)
        self.bot3 = DoubleConv(512, 256)

        self.up1 = UpScaler(512, 128)
        self.sa4 = SelfAttention(128, 16)
        self.up2 = UpScaler(256, 64)
        self.sa5 = SelfAttention(64, 32)
        self.up3 = UpScaler(128, 64)
        self.sa6 = SelfAttention(64, 64)
        self.outc = nn.Conv2d(64, channel_out, kernel_size=1)


        #giving the capacity to train with or without classes or labels
        if num_classes is not None:
            self.label_emb = nn.Embedding(num_classes, time_dim)

    def pos_encoding(self, t, channels):
        inv_freq = 1.0 / (
            10000
            ** (torch.arange(0, channels, 2, device=self.device).float() / channels)
        )
        pos_enc_a = torch.sin(t.repeat(1, channels // 2) * inv_freq)
        pos_enc_b = torch.cos(t.repeat(1, channels // 2) * inv_freq)
        pos_enc = torch.cat([pos_enc_a, pos_enc_b], dim=-1)
        return pos_enc

    def forward(self, x, t, conditional_labels):

        # unsqueeze turns an n.d. tensor into an (n+1).d. one by adding an extra dimension of depth 1.
        #unsqueeze(-1) adds a new dimension after the last index,
        t = t.unsqueeze(-1).type(torch.float)
        t = self.pos_encoding(t, self.time_dim)

        if conditional_labels is not None:
            t += self.label_emb(conditional_labels)
            #adding embedded labels to positional encoded time steps

        x1 = self.inc(x)
        x2 = self.down1(x1, t)
        x2 = self.sa1(x2)
        x3 = self.down2(x2, t)
        x3 = self.sa2(x3)
        x4 = self.down3(x3, t)
        x4 = self.sa3(x4)

        x4 = self.bot1(x4)
        x4 = self.bot2(x4)
        x4 = self.bot3(x4)

        x = self.up1(x4, x3, t)
        x = self.sa4(x)
        x = self.up2(x, x2, t)
        x = self.sa5(x)
        x = self.up3(x, x1, t)
        x = self.sa6(x)
        output = self.outc(x)
        return output
    


# two convolutaional layers
# 
class DoubleConv(nn.Module):

    """(convolution => [BN] => ReLU) * 2"""
    """This was a RElu and BatchNorm i nthe original UNET CODE"""
    def __init__(self, in_channels, out_channels, mid_channels=None, residual=False):
        super().__init__()

        self.residual = residual

        if not mid_channels:
            mid_channels = out_channels

        self.double_conv = nn.Sequential(
            nn.Conv2d(in_channels, mid_channels, kernel_size=3, padding=1, bias=False),
            nn.GroupNorm(1, mid_channels),
            nn.GELU(), #Gaussian Error Linear Units En lugar del RELU tradicional
            nn.Conv2d(mid_channels, out_channels, kernel_size=3, padding=1, bias=False),
            nn.GroupNorm(1, out_channels),
        )

    def forward(self, x):
        if self.residual:
            return F.gelu(x + self.double_conv(x))
        else:
            return self.double_conv(x)
        
class DownScaler(nn.Module):
    def __init__(self, in_channels, out_channels, emb_dim=256):
        super().__init__()
        self.maxpool_conv = nn.Sequential(
            nn.MaxPool2d(2),
            DoubleConv(in_channels, in_channels, residual=True),
            DoubleConv(in_channels, out_channels),
        )

        self.emb_layer = nn.Sequential(
            nn.SiLU(),
            nn.Linear(
                emb_dim,
                out_channels
            ),
        )

    def forward(self, x, t):

        x = self.maxpool_conv(x)

        #We are embedding the proper time dimension to the proper dimension
        emb = self.emb_layer(t)[:, :, None, None].repeat(1, 1, x.shape[-2], x.shape[-1])
        #we add both together. the input and the time positional encoding projection.
        return x + emb


class UpScaler(nn.Module):
    def __init__(self, in_channels, out_channels, emb_dim=256):
        super().__init__()

        self.up = nn.Upsample(scale_factor=2, mode="bilinear", align_corners=True)
        self.conv = nn.Sequential(
            DoubleConv(in_channels, in_channels, residual=True),
            DoubleConv(in_channels, out_channels, in_channels // 2),
        )

        self.emb_layer = nn.Sequential(
            nn.SiLU(),
            nn.Linear(
                emb_dim,
                out_channels
            ),
        )

    def forward(self, x, skip_x, t):
        x = self.up(x)

        x = torch.cat([skip_x, x], dim=1)

        x = self.conv(x)
        emb = self.emb_layer(t)[:, :, None, None].repeat(1, 1, x.shape[-2], x.shape[-1])
        return x + emb
    

class SelfAttention(nn.Module):
    def __init__(self, channels, size):
        super(SelfAttention, self).__init__()

        self.channels = channels
        self.size = size

        self.mha = nn.MultiheadAttention(channels, 4, batch_first=True) #embed dimension, numer of heads
        self.ln = nn.LayerNorm([channels])
        self.ff_self = nn.Sequential(
            nn.LayerNorm([channels]),
            nn.Linear(channels, channels),
            nn.GELU(),
            nn.Linear(channels, channels),
        )

    def forward(self, x):

        x = x.view(-1, self.channels, self.size * self.size).swapaxes(1, 2) #bringing images into the right shape

        x_ln = self.ln(x)
        attention_value, _ = self.mha(x_ln, x_ln, x_ln) #Query, Keys, Values The otput could be attention and wights
        attention_value = attention_value + x
        attention_value = self.ff_self(attention_value) + attention_value
        return attention_value.swapaxes(2, 1).view(-1, self.channels, self.size, self.size) #flattening and making channel last




class VAE(nn.Module):

    def __init__(self, input_dim=(64*64*3), hidden_dim=400, latent_dim=200, device="cpu"):
       
        super(VAE, self).__init__()
        self.device=device
        # encoder
        self.encoder = nn.Sequential(
            nn.Linear(input_dim, hidden_dim),
            nn.LeakyReLU(0.2),
            nn.Linear(hidden_dim, latent_dim),
            nn.LeakyReLU(0.2)
            )
        
        # latent mean and variance 
        self.mean_layer = nn.Linear(latent_dim, 2)
        self.logvar_layer = nn.Linear(latent_dim, 2)
        
        # decoder
        self.decoder = nn.Sequential(
            nn.Linear(2, latent_dim),
            nn.LeakyReLU(0.2),
            nn.Linear(latent_dim, hidden_dim),
            nn.LeakyReLU(0.2),
            nn.Linear(hidden_dim, input_dim),
            nn.Sigmoid()
            )
     
    def encode(self, x):
        x = self.encoder(x)
        mean, logvar = self.mean_layer(x), self.logvar_layer(x)
        return mean, logvar

    def reparameterization(self, mean, var):
        epsilon = torch.randn_like(var).to(self.device)      
        z = mean + var*epsilon
        return z

    def decode(self, x):
        return self.decoder(x)

        # def forward(self, x):
        #     mean, log_var = self.encode(x)
        #     z = self.reparameterization(mean, log_var)
        #     x_hat = self.decode(z)
        #     return x_hat, mean, log_var
        
    def forward(self, x):
        mean, log_var = self.encode(x)
        z = self.reparameterization(mean, torch.exp(0.5 * log_var)) 
        x_hat = self.decode(z)  
        return x_hat, mean, log_var
    

"""This section are building blocks of Stable diffusion"""
"""Corresponds to stable diffusion Autoencoder"""

class Autoencoder(nn.Module) :
    def __init__(self, encoder: 'Encoder', decoder: 'Decoder', emb_channels: int, z_channels: int):
        super().__init__()
        self.encoder = encoder
        self.decoder = decoder

        #emb_channels is the number of dimensions in the quantized embedding space
        #z_channels is the number of channels in the embedding space


        self.quant_conv = nn.Conv2d(2 * z_channels, 2 * emb_channels, 1)
        self.post_quant_conv = nn.Conv2d(emb_channels, z_channels, 1) 

    def encode(self, img: torch.Tensor)->"GaussianDistribution":
        #he image tensor with shape [batch_size, img_channels, img_height, img_width
        z = self.encoder(img) #embeddings with shape [batch_size, z_channels * 2, z_height, z_height]

        moments = self.quant_conv(z) #Get the moments in the quantized embedding space
        
        return moments, GaussianDistribution(moments)
    
    #Decode images from latent representation

    def decode(self, z: torch.Tensor):
        #z is the latent representation with shape [batch_size, emb_channels, z_height, z_height]
        z = self.post_quant_conv(z)
        return self.decoder(z)

class Encoder(nn.Module):
    """
    ## Encoder module
    """

    def __init__(self, *, channels: int, channel_multipliers: List[int], n_resnet_blocks: int,
                 in_channels: int, z_channels: int):
        """
        :param channels: is the number of channels in the first convolution layer
        :param channel_multipliers: are the multiplicative factors for the number of channels in the
            subsequent blocks
        :param n_resnet_blocks: is the number of resnet layers at each resolution
        :param in_channels: is the number of channels in the image
        :param z_channels: is the number of channels in the embedding space
        """
        super().__init__()

        # Number of blocks of different resolutions.
        # The resolution is halved at the end each top level block
        n_resolutions = len(channel_multipliers)

        # Initial $3 \times 3$ convolution layer that maps the image to `channels`
        self.conv_in = nn.Conv2d(in_channels, channels, 3, stride=1, padding=1)

        # Number of channels in each top level block
        channels_list = [m * channels for m in [1] + channel_multipliers]
        print(channels_list)
        # List of top-level blocks
        self.down = nn.ModuleList()
        # Create top-level blocks
        for i in range(n_resolutions):
            # Each top level block consists of multiple ResNet Blocks and down-sampling
            resnet_blocks = nn.ModuleList()
            # Add ResNet Blocks
            for _ in range(n_resnet_blocks):
                resnet_blocks.append(ResnetBlock(channels, channels_list[i + 1]))
                channels = channels_list[i + 1]
            # Top-level block
            down = nn.Module()
            down.block = resnet_blocks
            # Down-sampling at the end of each top level block except the last
            if i != n_resolutions - 1:
                down.downsample = DownSample(channels)
            else:
                down.downsample = nn.Identity()
            #
            self.down.append(down)

        # Final ResNet blocks with attention
        self.mid = nn.Module()
        self.mid.block_1 = ResnetBlock(channels, channels)
        self.mid.attn_1 = AttnBlock(channels)
        self.mid.block_2 = ResnetBlock(channels, channels)

        # Map to embedding space with a $3 \times 3$ convolution
        self.norm_out = normalization(channels)
        self.conv_out = nn.Conv2d(channels, 2 * z_channels, 3, stride=1, padding=1)

    def forward(self, img: torch.Tensor):
        """
        :param img: is the image tensor with shape `[batch_size, img_channels, img_height, img_width]`
        """

        # Map to `channels` with the initial convolution
        x = self.conv_in(img)

        # Top-level blocks
        for down in self.down:
            # ResNet Blocks
            for block in down.block:
                x = block(x)
            # Down-sampling
            x = down.downsample(x)

        # Final ResNet blocks with attention
        x = self.mid.block_1(x)
        x = self.mid.attn_1(x)
        x = self.mid.block_2(x)

        # Normalize and map to embedding space
        x = self.norm_out(x)
        x = swish(x)
        x = self.conv_out(x)

        #
        return x


class Decoder(nn.Module):
    """
    ## Decoder module
    """

    def __init__(self, *, channels: int, channel_multipliers: List[int], n_resnet_blocks: int,
                 out_channels: int, z_channels: int):
        """
        :param channels: is the number of channels in the final convolution layer
        :param channel_multipliers: are the multiplicative factors for the number of channels in the
            previous blocks, in reverse order
        :param n_resnet_blocks: is the number of resnet layers at each resolution
        :param out_channels: is the number of channels in the image
        :param z_channels: is the number of channels in the embedding space
        """
        super().__init__()

        # Number of blocks of different resolutions.
        # The resolution is halved at the end each top level block
        num_resolutions = len(channel_multipliers)

        # Number of channels in each top level block, in the reverse order
        channels_list = [m * channels for m in channel_multipliers]
        # Number of channels in the  top-level block
        channels = channels_list[-1]

        # Initial $3 \times 3$ convolution layer that maps the embedding space to `channels`
        self.conv_in = nn.Conv2d(z_channels, channels, 3, stride=1, padding=1)

        # ResNet blocks with attention
        self.mid = nn.Module()
        self.mid.block_1 = ResnetBlock(channels, channels)
        self.mid.attn_1 = AttnBlock(channels)
        self.mid.block_2 = ResnetBlock(channels, channels)

        # List of top-level blocks
        self.up = nn.ModuleList()
        # Create top-level blocks
        for i in reversed(range(num_resolutions)):
            # Each top level block consists of multiple ResNet Blocks and up-sampling
            resnet_blocks = nn.ModuleList()
            # Add ResNet Blocks
            for _ in range(n_resnet_blocks + 1):
                resnet_blocks.append(ResnetBlock(channels, channels_list[i]))
                channels = channels_list[i]
            # Top-level block
            up = nn.Module()
            up.block = resnet_blocks
            # Up-sampling at the end of each top level block except the first
            if i != 0:
                up.upsample = UpSample(channels)
            else:
                up.upsample = nn.Identity()
            # Prepend to be consistent with the checkpoint
            self.up.insert(0, up)

        # Map to image space with a $3 \times 3$ convolution
        self.norm_out = normalization(channels)
        self.conv_out = nn.Conv2d(channels, out_channels, 3, stride=1, padding=1)

    def forward(self, z: torch.Tensor):
        """
        :param z: is the embedding tensor with shape `[batch_size, z_channels, z_height, z_height]`
        """

        # Map to `channels` with the initial convolution
        h = self.conv_in(z)

        # ResNet blocks with attention
        h = self.mid.block_1(h)
        h = self.mid.attn_1(h)
        h = self.mid.block_2(h)

        # Top-level blocks
        for up in reversed(self.up):
            # ResNet Blocks
            for block in up.block:
                h = block(h)
            # Up-sampling
            h = up.upsample(h)

        # Normalize and map to image space
        h = self.norm_out(h)
        h = swish(h)
        img = self.conv_out(h)

        #
        return img


class GaussianDistribution:
    """
    ## Gaussian Distribution
    """

    def __init__(self, parameters: torch.Tensor):
        """
        :param parameters: are the means and log of variances of the embedding of shape
            `[batch_size, z_channels * 2, z_height, z_height]`
        """
        # Split mean and log of variance
        self.mean, log_var = torch.chunk(parameters, 2, dim=1)
        # Clamp the log of variances
        self.log_var = torch.clamp(log_var, -30.0, 20.0)
        # Calculate standard deviation
        self.std = torch.exp(0.5 * self.log_var)

    def sample(self):
        # Sample from the distribution
        return self.mean + self.std * torch.randn_like(self.std)


class AttnBlock(nn.Module):
    """
    ## Attention block
    """

    def __init__(self, channels: int):
        """
        :param channels: is the number of channels
        """
        super().__init__()
        # Group normalization
        self.norm = normalization(channels)
        # Query, key and value mappings
        self.q = nn.Conv2d(channels, channels, 1)
        self.k = nn.Conv2d(channels, channels, 1)
        self.v = nn.Conv2d(channels, channels, 1)
        # Final $1 \times 1$ convolution layer
        self.proj_out = nn.Conv2d(channels, channels, 1)
        # Attention scaling factor
        self.scale = channels ** -0.5

    def forward(self, x: torch.Tensor):
        """
        :param x: is the tensor of shape `[batch_size, channels, height, width]`
        """
        # Normalize `x`
        x_norm = self.norm(x)
        # Get query, key and vector embeddings
        q = self.q(x_norm)
        k = self.k(x_norm)
        v = self.v(x_norm)

        # Reshape to query, key and vector embeedings from
        # `[batch_size, channels, height, width]` to
        # `[batch_size, channels, height * width]`
        b, c, h, w = q.shape
        q = q.view(b, c, h * w)
        k = k.view(b, c, h * w)
        v = v.view(b, c, h * w)

        # Compute $\underset{seq}{softmax}\Bigg(\frac{Q K^\top}{\sqrt{d_{key}}}\Bigg)$
        #la formula de Atenci[on]
         # batch matrix multiplication
        # As = torch.randn(3, 2, 5)
        # Bs = torch.randn(3, 5, 4)
        # torch.einsum('bij,bjk->bik', As, Bs)
        #This contracts along the channel dimmension
        attn = torch.einsum('bci,bcj->bij', q, k) * self.scale
        attn = F.softmax(attn, dim=2)



        # Compute $\underset{seq}{softmax}\Bigg(\frac{Q K^\top}{\sqrt{d_{key}}}\Bigg)V$
        #multiplicar V a la parte con el softmax de Q y K

        out = torch.einsum('bij,bcj->bci', attn, v)

        # Reshape back to `[batch_size, channels, height, width]`
        #this can be done as de 'i' dimension is in fact h*w
        out = out.view(b, c, h, w)

        # Final $1 \times 1$ convolution layer
        out = self.proj_out(out)

        # Add residual connection
        return x + out


class UpSample(nn.Module):
    """
    ## Up-sampling layer
    """
    def __init__(self, channels: int):
        """
        :param channels: is the number of channels
        """
        super().__init__()
        # $3 \times 3$ convolution mapping
        self.conv = nn.Conv2d(channels, channels, 3, padding=1)

    def forward(self, x: torch.Tensor):
        """
        :param x: is the input feature map with shape `[batch_size, channels, height, width]`
        """
        # Up-sample by a factor of $2$
        x = F.interpolate(x, scale_factor=2.0, mode="nearest")
        # Apply convolution
        return self.conv(x)


class DownSample(nn.Module):
    """
    ## Down-sampling layer
    """
    def __init__(self, channels: int):
        """
        :param channels: is the number of channels
        """
        super().__init__()
        # $3 \times 3$ convolution with stride length of $2$ to down-sample by a factor of $2$
        self.conv = nn.Conv2d(channels, channels, 3, stride=2, padding=0)

    def forward(self, x: torch.Tensor):
        """
        :param x: is the input feature map with shape `[batch_size, channels, height, width]`
        """
        # Add padding
        x = F.pad(x, (0, 1, 0, 1), mode="constant", value=0)
        # Apply convolution
        return self.conv(x)


class ResnetBlock(nn.Module):
    """
    ## ResNet Block
    """
    def __init__(self, in_channels: int, out_channels: int):
        """
        :param in_channels: is the number of channels in the input
        :param out_channels: is the number of channels in the output
        """
        super().__init__()
        # First normalization and convolution layer
        self.norm1 = normalization(in_channels)
        self.conv1 = nn.Conv2d(in_channels, out_channels, 3, stride=1, padding=1)
        # Second normalization and convolution layer
        self.norm2 = normalization(out_channels)
        self.conv2 = nn.Conv2d(out_channels, out_channels, 3, stride=1, padding=1)
        # `in_channels` to `out_channels` mapping layer for residual connection
        if in_channels != out_channels:
            self.nin_shortcut = nn.Conv2d(in_channels, out_channels, 1, stride=1, padding=0)
        else:
            self.nin_shortcut = nn.Identity()

    def forward(self, x: torch.Tensor):
        """
        :param x: is the input feature map with shape `[batch_size, channels, height, width]`
        """

        h = x

        # First normalization and convolution layer
        h = self.norm1(h)
        h = swish(h)
        h = self.conv1(h)

        # Second normalization and convolution layer
        h = self.norm2(h)
        h = swish(h)
        h = self.conv2(h)

        # Map and add residual
        return self.nin_shortcut(x) + h


def swish(x: torch.Tensor):
    """
    ### Swish activation

    $$x \cdot \sigma(x)$$
    """
    return x * torch.sigmoid(x)


def normalization(channels: int):
    """
    ### Group normalization

    This is a helper function, with fixed number of groups and `eps`.
    """
    return nn.GroupNorm(num_groups=32, num_channels=channels, eps=1e-6)


#Cliip embedder for the conditioning stage
class CLIPTextEmbedder(nn.Module):
    """
    ## CLIP Text Embedder
    """

    def __init__(self, version: str = "openai/clip-vit-large-patch14", device="cuda:0", max_length: int = 77):
        """
        :param version: is the model version
        :param device: is the device
        :param max_length: is the max length of the tokenized prompt
        """
        super().__init__()
        # Load the tokenizer
        self.tokenizer = CLIPTokenizer.from_pretrained(version)
        # Load the CLIP transformer
        self.transformer = CLIPTextModel.from_pretrained(version).eval()

        self.device = device
        self.max_length = max_length

    def forward(self, prompts: List[str]):
        """
        :param prompts: are the list of prompts to embed
        """
        # Tokenize the prompts
        batch_encoding = self.tokenizer(prompts, truncation=True, max_length=self.max_length, return_length=True,
                                        return_overflowing_tokens=False, padding="max_length", return_tensors="pt")
        # Get token ids
        tokens = batch_encoding["input_ids"].to(self.device)
        # Get CLIP embeddings
        return self.transformer(input_ids=tokens).last_hidden_state
    
"""Stable diffusion Unet"""

class UNetModel(nn.Module):
    """
    ## U-Net model
    """

    def __init__(
            self, *,
            in_channels: int,
            out_channels: int,
            channels: int,
            n_res_blocks: int,
            attention_levels: List[int],
            channel_multipliers: List[int],
            n_heads: int,
            tf_layers: int = 1,
            d_cond: int = 768):
        """
        :param in_channels: is the number of channels in the input feature map
        :param out_channels: is the number of channels in the output feature map
        :param channels: is the base channel count for the model
        :param n_res_blocks: number of residual blocks at each level
        :param attention_levels: are the levels at which attention should be performed
        :param channel_multipliers: are the multiplicative factors for number of channels for each level
        :param n_heads: is the number of attention heads in the transformers
        :param tf_layers: is the number of transformer layers in the transformers
        :param d_cond: is the size of the conditional embedding in the transformers
        """
        super().__init__()
        self.channels = channels

        # Number of levels
        levels = len(channel_multipliers)
        # Size time embeddings
        d_time_emb = channels * 4
        self.time_embed = nn.Sequential(
            nn.Linear(channels, d_time_emb),
            nn.SiLU(),
            nn.Linear(d_time_emb, d_time_emb),
        )

        # Input half of the U-Net
        self.input_blocks = nn.ModuleList()
        # Initial $3 \times 3$ convolution that maps the input to `channels`.
        # The blocks are wrapped in `TimestepEmbedSequential` module because
        # different modules have different forward function signatures;
        # for example, convolution only accepts the feature map and
        # residual blocks accept the feature map and time embedding.
        # `TimestepEmbedSequential` calls them accordingly.
        self.input_blocks.append(TimestepEmbedSequential(
            nn.Conv2d(in_channels, channels, 3, padding=1)))
        # Number of channels at each block in the input half of U-Net
        input_block_channels = [channels]
        # Number of channels at each level
        channels_list = [channels * m for m in channel_multipliers]
        # Prepare levels
        for i in range(levels):
            # Add the residual blocks and attentions
            for _ in range(n_res_blocks):
                # Residual block maps from previous number of channels to the number of
                # channels in the current level
                layers = [ResBlock(channels, d_time_emb, out_channels=channels_list[i])]
                channels = channels_list[i]
                # Add transformer
                if i in attention_levels:
                    layers.append(SpatialTransformer(channels, n_heads, tf_layers, d_cond))
                # Add them to the input half of the U-Net and keep track of the number of channels of
                # its output
                self.input_blocks.append(TimestepEmbedSequential(*layers))
                input_block_channels.append(channels)
            # Down sample at all levels except last
            if i != levels - 1:
                self.input_blocks.append(TimestepEmbedSequential(DownSample(channels)))
                input_block_channels.append(channels)

        # The middle of the U-Net
        self.middle_block = TimestepEmbedSequential(
            ResBlock(channels, d_time_emb),
            SpatialTransformer(channels, n_heads, tf_layers, d_cond),
            ResBlock(channels, d_time_emb),
        )

        # Second half of the U-Net
        self.output_blocks = nn.ModuleList([])
        # Prepare levels in reverse order
        for i in reversed(range(levels)):
            # Add the residual blocks and attentions
            for j in range(n_res_blocks + 1):
                # Residual block maps from previous number of channels plus the
                # skip connections from the input half of U-Net to the number of
                # channels in the current level.
                layers = [ResBlock(channels + input_block_channels.pop(), d_time_emb, out_channels=channels_list[i])]
                channels = channels_list[i]
                # Add transformer
                if i in attention_levels:
                    layers.append(SpatialTransformer(channels, n_heads, tf_layers, d_cond))
                # Up-sample at every level after last residual block
                # except the last one.
                # Note that we are iterating in reverse; i.e. `i == 0` is the last.
                if i != 0 and j == n_res_blocks:
                    layers.append(UpSample(channels))
                # Add to the output half of the U-Net
                self.output_blocks.append(TimestepEmbedSequential(*layers))

        # Final normalization and $3 \times 3$ convolution
        self.out = nn.Sequential(
            normalization(channels),
            nn.SiLU(),
            nn.Conv2d(channels, out_channels, 3, padding=1),
        )

    def time_step_embedding(self, time_steps: torch.Tensor, max_period: int = 10000):
        """
        ## Create sinusoidal time step embeddings

        :param time_steps: are the time steps of shape `[batch_size]`
        :param max_period: controls the minimum frequency of the embeddings.
        """
        # $\frac{c}{2}$; half the channels are sin and the other half is cos,
        half = self.channels // 2
        # $\frac{1}{10000^{\frac{2i}{c}}}$
        frequencies = torch.exp(
            -math.log(max_period) * torch.arange(start=0, end=half, dtype=torch.float32) / half
        ).to(device=time_steps.device)
        # $\frac{t}{10000^{\frac{2i}{c}}}$
        args = time_steps[:, None].float() * frequencies[None]
        # $\cos\Bigg(\frac{t}{10000^{\frac{2i}{c}}}\Bigg)$ and $\sin\Bigg(\frac{t}{10000^{\frac{2i}{c}}}\Bigg)$
        return torch.cat([torch.cos(args), torch.sin(args)], dim=-1)

    def forward(self, x: torch.Tensor, time_steps: torch.Tensor, cond: torch.Tensor):
        """
        :param x: is the input feature map of shape `[batch_size, channels, width, height]`
        :param time_steps: are the time steps of shape `[batch_size]`
        :param cond: conditioning of shape `[batch_size, n_cond, d_cond]`
        """
        # To store the input half outputs for skip connections
        x_input_block = []

        # Get time step embeddings
        t_emb = self.time_step_embedding(time_steps)
        t_emb = self.time_embed(t_emb)

        # Input half of the U-Net
        for module in self.input_blocks:
            x = module(x, t_emb, cond)
            x_input_block.append(x)
        # Middle of the U-Net
        x = self.middle_block(x, t_emb, cond)
        # Output half of the U-Net
        for module in self.output_blocks:
            x = torch.cat([x, x_input_block.pop()], dim=1)
            x = module(x, t_emb, cond)

        # Final normalization and $3 \times 3$ convolution
        return self.out(x)


class TimestepEmbedSequential(nn.Sequential):
    """
    ### Sequential block for modules with different inputs

    This sequential module can compose of different modules such as `ResBlock`,
    `nn.Conv` and `SpatialTransformer` and calls them with the matching signatures
    """

    def forward(self, x, t_emb, cond=None):
        for layer in self:
            if isinstance(layer, ResBlock):
                x = layer(x, t_emb)
            elif isinstance(layer, SpatialTransformer):
                x = layer(x, cond)
            else:
                x = layer(x)
        return x


class UpSample(nn.Module):
    """
    ### Up-sampling layer
    """

    def __init__(self, channels: int):
        """
        :param channels: is the number of channels
        """
        super().__init__()
        # $3 \times 3$ convolution mapping
        self.conv = nn.Conv2d(channels, channels, 3, padding=1)

    def forward(self, x: torch.Tensor):
        """
        :param x: is the input feature map with shape `[batch_size, channels, height, width]`
        """
        # Up-sample by a factor of $2$
        x = F.interpolate(x, scale_factor=2, mode="nearest")
        # Apply convolution
        return self.conv(x)


class DownSample(nn.Module):
    """
    ## Down-sampling layer
    """

    def __init__(self, channels: int):
        """
        :param channels: is the number of channels
        """
        super().__init__()
        # $3 \times 3$ convolution with stride length of $2$ to down-sample by a factor of $2$
        self.op = nn.Conv2d(channels, channels, 3, stride=2, padding=1)

    def forward(self, x: torch.Tensor):
        """
        :param x: is the input feature map with shape `[batch_size, channels, height, width]`
        """
        # Apply convolution
        return self.op(x)


class ResBlock(nn.Module):
    """
    ## ResNet Block
    """

    def __init__(self, channels: int, d_t_emb: int, *, out_channels=None):
        """
        :param channels: the number of input channels
        :param d_t_emb: the size of timestep embeddings
        :param out_channels: is the number of out channels. defaults to `channels.
        """
        super().__init__()
        # `out_channels` not specified
        if out_channels is None:
            out_channels = channels

        # First normalization and convolution
        self.in_layers = nn.Sequential(
            normalization(channels),
            nn.SiLU(),
            nn.Conv2d(channels, out_channels, 3, padding=1),
        )

        # Time step embeddings
        self.emb_layers = nn.Sequential(
            nn.SiLU(),
            nn.Linear(d_t_emb, out_channels),
        )
        # Final convolution layer
        self.out_layers = nn.Sequential(
            normalization(out_channels),
            nn.SiLU(),
            nn.Dropout(0.),
            nn.Conv2d(out_channels, out_channels, 3, padding=1)
        )

        # `channels` to `out_channels` mapping layer for residual connection
        if out_channels == channels:
            self.skip_connection = nn.Identity()
        else:
            self.skip_connection = nn.Conv2d(channels, out_channels, 1)

    def forward(self, x: torch.Tensor, t_emb: torch.Tensor):
        """
        :param x: is the input feature map with shape `[batch_size, channels, height, width]`
        :param t_emb: is the time step embeddings of shape `[batch_size, d_t_emb]`
        """
        # Initial convolution
        h = self.in_layers(x)
        # Time step embeddings
        t_emb = self.emb_layers(t_emb).type(h.dtype)
        # Add time step embeddings
        h = h + t_emb[:, :, None, None]
        # Final convolution
        h = self.out_layers(h)
        # Add skip connection
        return self.skip_connection(x) + h


class GroupNorm32(nn.GroupNorm):
    """
    ### Group normalization with float32 casting
    """

    def forward(self, x):
        return super().forward(x.float()).type(x.dtype)


def normalization(channels):
    """
    ### Group normalization

    This is a helper function, with fixed number of groups..
    """
    return GroupNorm32(32, channels)


def _test_time_embeddings():
    """
    Test sinusoidal time step embeddings
    """
    import matplotlib.pyplot as plt

    plt.figure(figsize=(15, 5))
    m = UNetModel(in_channels=1, out_channels=1, channels=320, n_res_blocks=1, attention_levels=[],
                  channel_multipliers=[],
                  n_heads=1, tf_layers=1, d_cond=1)
    te = m.time_step_embedding(torch.arange(0, 1000))
    plt.plot(np.arange(1000), te[:, [50, 100, 190, 260]].numpy())
    plt.legend(["dim %d" % p for p in [50, 100, 190, 260]])
    plt.title("Time embeddings")
    plt.show()


#


import os
import torch
import torchvision
from PIL import Image
from matplotlib import pyplot as plt
from torch.utils.data import DataLoader

import numpy as np
import cv2 as cv
import os
import glob
import ezdxf
import imutils

from ..DataManager import datamanager

    

def plot_images(images):

    
    plt.figure(figsize=(16, 16))
    plt.imshow(torch.cat([
        torch.cat([i for i in images.cpu()], dim=-1),
    ], dim=-2).permute(1, 2, 0).cpu())
    plt.show()


def save_images(images, path, **kwargs):
    grid = torchvision.utils.make_grid(images, **kwargs)
    ndarr = grid.permute(1, 2, 0).to('cpu').numpy()
    im = Image.fromarray(ndarr)
    im.save(path)



""""Converting an image to tensor and normalizing"""
def get_data(image_size,resize, dataset_path,batch_size):

    transforms = torchvision.transforms.Compose([
        torchvision.transforms.Resize(resize),  # args.image_size + 1/4 *args.image_size
        torchvision.transforms.RandomResizedCrop(image_size, scale=(0.9, 0.9)),
        torchvision.transforms.ToTensor(),
        torchvision.transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
    ])
    dataset = torchvision.datasets.ImageFolder(dataset_path, transform=transforms)
    dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=False)
    #returns a data loader with normalized images

    return dataloader, dataset.imgs

""""Converting an image to tensor and normalizing"""
def get_data_with_labels(image_size,resize, randomResize, dataset_path,batch_size, drop_last,filter):

    transforms = torchvision.transforms.Compose([
        #torchvision.transforms.Resize(resize),  # args.image_size + 1/4 *args.image_size
        torchvision.transforms.RandomResizedCrop(image_size, scale=(randomResize, randomResize)),
        torchvision.transforms.ToTensor(),
        torchvision.transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
    ])

    cursomDataset = datamanager.CustomDataset(dataset_path,transforms,filter)

    # if not (filter is None):

    #     trainset_1 = torch.utils.data.Subset(cursomDataset, self.filter)
    #     dataloader = DataLoader(trainset_1,  batch_size=batch_size, shuffle=True, drop_last=drop_last)
    # else:
    dataloader = DataLoader(cursomDataset, batch_size=batch_size, shuffle=True, drop_last=drop_last)

    #returns a data loader with normalized images

    return dataloader

def get_data_denormalize(image_size,resize, dataset_path,batch_size):

    transforms = torchvision.transforms.Compose([
        torchvision.transforms.Resize(resize),  # args.image_size + 1/4 *args.image_size
        torchvision.transforms.RandomResizedCrop(image_size, scale=(0.8, 1.0)),
        torchvision.transforms.ToTensor(),

    ])

    #Ojo realmente si se necesita normalizar transforms.Normalize((0.1307,), (0.3081,))
    dataset = torchvision.datasets.ImageFolder(dataset_path, transform=transforms)
    dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=True)

    #for inputs, targets in dataloader:
    #    print(inputs.size())
    #    print(targets.size())

    #returns a data loader with normalized images

    return dataloader


def setup_logging(run_name):
    os.makedirs("models", exist_ok=True)
    os.makedirs("results", exist_ok=True)
    os.makedirs(os.path.join("models", run_name), exist_ok=True)
    os.makedirs(os.path.join("results", run_name), exist_ok=True)

def load_images(path):
    loadedImages = []
    # return array of images
    filenames = glob.glob(path)
    filenames.sort()
    for imgdata in filenames:
        # determine whether it is an image.
        if os.path.isfile(os.path.splitext(os.path.join(path, imgdata))[0] + ".png"):
            img_array = cv2.imread(os.path.join(path, imgdata))
            img_array = np.float32(img_array)
            img_array = cv2.cvtColor(img_array, cv2.COLOR_BGR2GRAY)
            loadedImages.append(img_array)
    return loadedImages

class Binary:
    
    def convert(self,results_folder):
        # set folder of images
        folder = results_folder  # Set 
        path = folder+'*-bw.png'
        filenames = glob.glob(path)
        filenames.sort()
        imgs = load_images(path) # load images
        imgs = np.asarray(imgs)
        print(np.shape(imgs))
        
        for i in range(len(filenames)):
            basename = os.path.basename(filenames[i])
            # Turn image array into binary array (black to 1, white to 0)
            binary = np.zeros(shape = (np.shape(imgs)[1], np.shape(imgs)[2]), dtype = np.uint8)
            img = imgs[i][:][:]
            print(np.shape(binary))
            print(np.shape(img[:][:][0]))
            binary[img[:][:] <= 50] = 1
            
            print(len(binary[binary==1]))
            # doubling the amount of pixels in both dimensions
            resize_fac = 2
            height, width = imgs.shape[:][:][1:]
            new = np.ones(shape = (resize_fac*height, resize_fac*width), dtype = np.uint8)
            print(np.shape(new))
            print(np.shape(binary))
            new[:binary.shape[0],:binary.shape[1]] = binary[:][:]
            for i in range(height-1,-1,-1):
                for j in range(width-1,-1,-1):
                    cur = new[i][j]
                    new[resize_fac*i:resize_fac*(i+1),resize_fac*j:resize_fac*(j+1)] = [[cur]*resize_fac] * resize_fac
                        
            print(len(new[new==1]))
            print(len(new[new==1])/(resize_fac**2) == len(binary[binary==1]))
            
            print(folder+basename[:-4]+'.txt')
            file1 = open(folder+basename[:-4]+'.txt', 'w')
            header = str(height*resize_fac)+' 1 '+str(height*resize_fac)+' \n'+str(width*resize_fac)+' 1 '+str(width*resize_fac)+' \n2 1 2\n'
            file1.write(header)
            body = new.reshape(new.shape[0]*new.shape[1],1)
            cnt=0
            for item in body:
                file1.write("%s\n" % item[0])
                cnt +=1
            for item in body:
                file1.write("%s\n" % item[0])
            
            file1.close()

class CAD():
    def __init__(self, images_folder, destination_folder):
        super().__init__()
        self.images_folder=images_folder
        self.destination_folder=destination_folder
        self.contours_path="./contours/"
        self.output_path=destination_folder

        self.colorContourList=[]

        isExist = os.path.exists(self.contours_path)
        if not isExist:

            # Create a new directory because it does not exist
            os.makedirs(self.contours_path)
            print("The new directory is created!")

        isExist = os.path.exists(self.output_path)
        if not isExist:

            # Create a new directory because it does not exist
            os.makedirs(self.output_path)
            print("The new directory is created!")

    """These functions help to separate 
    contours based on HSV mapping of the image"""

    def colorContour(self,upperBound, lowerBound,image,epsilon_coeff, threshold_Value,contour_name):
        #Channel separation
    
        im = cv.imread(image,cv.IMREAD_UNCHANGED)


        im2 =im.copy()
        im22 = cv.cvtColor(im2,cv.COLOR_BGR2RGB )
        
        hsv = cv.cvtColor(im2, cv.COLOR_BGR2HSV)

        # define range of blue color in HSV
        lower_red = np.array(lowerBound)
        upper_red = np.array(upperBound)
        # Threshold the HSV image to get only blue colors
        mask = cv.inRange(hsv,lower_red,upper_red)

        # Apply threshold on s - use automatic threshold algorithm (use THRESH_OTSU).
        _, thresh = cv.threshold(mask, threshold_Value, 255, cv.THRESH_BINARY+cv.THRESH_OTSU)

        # Find contours
        red_cnts = cv.findContours(thresh, cv.RETR_TREE, cv.CHAIN_APPROX_NONE)
        red_cnts = imutils.grab_contours(red_cnts) 
        #red_cnts= max(red_cnts, key=cv.contourArea) 
        #Drawing
        canvas = np.zeros_like(im22)
        size = im.shape

        #operating on every contour line
        smoothened=[]

        for contour in red_cnts:
            
            epsilon = epsilon_coeff*cv.arcLength(contour,True)
            cnt_aprox = cv.approxPolyDP(contour,epsilon,True)
        
            smoothened.append(np.asarray(cnt_aprox, dtype=np.int32))
        
            #x,y = contour.T
            # Convert from numpy arrays to normal arrays
            #x = x.tolist()[0]
            #y = y.tolist()[0]
            
            #if len(x)>2:
            
                # https://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.interpolate.splprep.html
                #tck, u = splprep([x,y], u=None, s=1.0, per=2,k=1)
                # https://docs.scipy.org/doc/numpy-1.10.1/reference/generated/numpy.linspace.html
                #u_new = np.linspace(u.min(), u.max(), 250)
                # https://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.interpolate.splev.html
                #x_new, y_new = splev(u_new, tck, der=0)
                # Convert it back to numpy format for opencv to be able to display it
                #res_array = [[[int(i[0]), int(i[1])]] for i in zip(x_new,y_new)]
                #smoothened.append(np.asarray(res_array, dtype=np.int32))
            cv.drawContours(canvas, smoothened,-1,(0, 255, 0), 1)

        cv.fillPoly(canvas, pts =smoothened, color=(255,255,255))

        #cv.fillPoly(canvas, pts =red_cnts, color=(255,255,255))

        cv.imwrite(self.contours_path+contour_name+".png", canvas, [cv.IMWRITE_PNG_COMPRESSION, 10]) 
        
        fig, ax = plt.subplots(1, 2, sharex='col', sharey='row',figsize=(5, 5))
        ax[0].imshow(im22)
        ax[1].imshow(canvas)
        
        return red_cnts,size

    def DXF_build(self,multiplier, GoalSize,currentSize,red_cnts, border_cnts,green_cnts, selectedLayer):


        
        dwg = ezdxf.new()#"AC1015"
        msp = dwg.modelspace()

        dwg.layers.new(name="conductor", dxfattribs={"color": 1})
        dwg.layers.new(name="dielectric", dxfattribs={"color": 8})
        dwg.layers.new(name="substrate", dxfattribs={"color": 5})

        """
        Warning! This is the drwaing unit, not the HFSS original units.
        To preserve the resolution everything needs to be rescaled and resized 
        if image size is 512x512 it means 512 nm or 512 mm dpending on the units declared in dxf file
        thus all drawing sizes must be scaled to fit.

        working units together with multipliers must be chosen based on the need to keep ratio and size as we import from HFSS
        """

        scale=multiplier*GoalSize/currentSize

        red_squeezed = [np.squeeze(cnt, axis=1) for cnt in red_cnts]
        inverted_red_squeezed = [scale*arr * [1, -1] for arr in red_squeezed]#*0.1

        green_squeezed = [np.squeeze(cnt, axis=1) for cnt in green_cnts]
        inverted_green_squeezed = [scale*arr * [1, -1] for arr in green_squeezed]#*0.1

        border_squeezed = [np.squeeze(cnt, axis=1) for cnt in border_cnts]
        inverted_border_squeezed = [scale*arr * [1, -1] for arr in border_squeezed]#*0.1

        """Select the layers to export"""
        

        for layer in selectedLayer:
            if layer=="conductor":

                for ctr in inverted_red_squeezed:
                    line=msp.add_lwpolyline(
                        ctr,
                        format="xyb", close=True,
                        dxfattribs={'layer': 'conductor'},
                    )
                    line.dxf.const_width = 0.5
            elif layer=="dielectric":

                for ctr in inverted_green_squeezed:
                    line=msp.add_lwpolyline(
                        ctr,
                        format="xyb", close=True,
                        dxfattribs={'layer': 'dielectric'},
                    )
                    line.dxf.const_width = 0.5
            elif layer=="substrate":

                for ctr in inverted_border_squeezed:
                    line=msp.add_lwpolyline(
                        ctr,
                        format="xyb", close=True,
                        dxfattribs={'layer': 'substrate'},
                    )
                    line.dxf.const_width = 0.5
            else:
                pass

            
            
            """To draw the outline """

        #for ctr in inverted_red_squeezed:
        #    for n in range(len(ctr)):
        #        if n >= len(ctr) - 1:
        #            n = 0
        #            
        #        try:
        #            msp.add_line(ctr[n], ctr[n + 1], dxfattribs={"layer": "red", "lineweight": 30})
        #        except IndexError:
        #            pass



        dwg.saveas(self.output_path+"output.dxf")

    def elevation_file(self,layers,**kwargs):
        
        units=kwargs["units"]
        SIMID=kwargs["simulation_id"]

        file = open(self.output_path+SIMID+'.tech', 'w')
        file.write('units"'+ units+ '"\n')

        for layer in layers:
            
            file.write(layer+" "+ layers[layer]["color"]+" "+ str(layers[layer]["zpos"]) +" "+ str(layers[layer]["thickness"]) +" "+'\n')

        file.close()